#
# Copyright (c) Daniel Sheffield 2021 - 2022
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from collections import OrderedDict
from decimal import Decimal, InvalidOperation
from typing import Callable, List, Union

import numpy as np
import urwid
from app.db_utils import QueryManager

from .. import COPYRIGHT
from ..widgets import (AutoCompleteEdit, AutoCompleteFloatEdit,
                       AutoCompletePopUp, FocusWidget, NoTabCheckBox)
from . import ActivityManager, show_or_exit


class PriceCheck(FocusWidget):

    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        
        if getattr(self.original_widget.original_widget, 'original_widget', None) is None:
            return super().keypress(size, key)
        
        if key == 'tab':
            self.advance_focus()
        elif key == 'shift tab':
            self.advance_focus(reverse=True)
        else:
            return super().keypress(size, key)

    def apply_choice(self, name, value):
        self.apply_changes(name, value)
        for k,v in self.data.items():
            if k == name or v:
                continue
            options = self.query_manager.unique_suggestions(k, **self.data)
            if len(options) == 1:
                self.apply_changes(k, list(options)[0])

    def apply_changes(self, name, value):
        if name in self.data:
            
            self.data.update({
                name: value,
            })
        self.update_historic_prices()

    def clear(self):
        for k in self.data:
            if k == 'organic':
                self.data[k] = 'mixed'
                continue
            self.data[k] = ''
        self.text_fields['rating'].set_text('')
        self.text_fields['spread'].set_text('')
        self.text_fields['marker'].set_text('')
        self.update()
        return self
    
    def update(self):
        for k in self.edit_fields:
            if self.data[k] != self.edit_fields[k].get_edit_text():
                self.edit_fields[k].set_edit_text(self.data[k])
        
        if self.data['organic'] != self.organic_checkbox.state:
            self.organic_checkbox.set_state(self.data['organic'])
        
        self.update_historic_prices()
        
        return self
    
    def update_rating(self, _avg, _min, _max, unit, price=None, quantity=None):
        if None in (_avg, _min, _max):
            return
        current = None if None in (price, quantity or None) else float(price/quantity)
        size = 14
        chars = ['|', *['-']*(size - 2), '|' ]
        rating = [' ']*len(chars)
        _min, _max = min(_min, current or _min), max(_max, current or _max)
        ls = np.linspace(_min, _max, len(chars))

        for idx, (e, a) in enumerate(zip(ls, ls[1:])):
            if e <= _avg < a:
                for c, (_idx,_) in zip('[{}]'.format(''.join(list(f'{_avg:>5.2f}')[:idx-3+5])), filter(lambda x: idx-3 < x[0], enumerate(chars))):
                    chars[_idx] = c
                chars[0] = '|'
                chars[-1] = '|'
            
            if current is not None and e <= current < a:
                rating[idx] = '^'
        
        if current is not None:
            self.text_fields['rating'].set_text(f"{current:>5.2f}/{unit} {current/_avg - 1:>4.0%}")
        
        if _min != _max:
            if current == _max:
                rating[-1] = '^'
            self.text_fields['spread'].set_text(f"{_min:>5.2f}{''.join(chars)}{_max:<5.2f}")
            self.text_fields['marker'].set_text(f"{' '*5}{''.join(rating)}{' '*5}")
        else:
            self.text_fields['spread'].set_text('')
            self.text_fields['marker'].set_text('')
        
    
    def update_historic_prices(self):
        organic = None if self.data['organic'] == 'mixed' else self.data['organic']
        sort = '$/unit' if self.buttons['sort_price'].state else 'ts'
        product, unit = self.data['product'] or None, self.data['unit'] or None
        try: 
            price = Decimal(self.data['price'])
        except InvalidOperation:
            price = None
        
        try:
            quantity = Decimal(self.data['quantity'])
        except InvalidOperation:
            quantity = None
        
        if None not in (sort, product, unit):
            self.text_fields['dbview'].set_text(
                self.query_manager.get_historic_prices(
                    lambda *args: self.update_rating(*args, unit, price=price, quantity=quantity),
                    sort, product, unit, organic=organic) 
            )

    def __init__(self,
        activity_manager: ActivityManager,
        query_manager: QueryManager,
        fields: List[str],
        top_pane, left_pane, right_pane, bottom_pane, badge,
        autocomplete_cb: Callable[[
            Union[AutoCompleteEdit, AutoCompleteFloatEdit], str, dict
        ], None],
    ):
        self.query_manager = query_manager
        self.top_pane = top_pane
        self.left_pane = left_pane
        self.right_pane = right_pane
        self.bottom_pane = bottom_pane
        self.badge = badge
        self.edit_fields = OrderedDict()
        self.text_fields = OrderedDict()
        self.data = OrderedDict()
        self.organic_checkbox = NoTabCheckBox(
            u"Organic",
            state='mixed',
        )
        urwid.connect_signal(self.organic_checkbox, 'change', lambda _,v: self.apply_changes('organic', v))
        
        fields = [f for f in fields ]
        for k in fields:
            if k in self.right_pane and k != 'unit':
                ef = AutoCompleteFloatEdit(('bg', k))
            elif k != 'organic':
                ef = AutoCompleteEdit(('bg', k))
            elif k == 'organic':
                self.data[k] = 'mixed'
                continue
            
            self.data[k] = ''
            urwid.connect_signal(ef, 'change', lambda w, v: self.apply_changes(w.name, v))
            urwid.connect_signal(ef, 'apply', lambda w, name: autocomplete_cb(w, name, self.data))
            self.edit_fields[k] = ef
        
        self.buttons = OrderedDict()
        group = []
        for k in top_pane:
            if k != 'sort':
                self.buttons.update({ k:  urwid.Button(('streak', f'{k.title()}')) })
                continue
            self.buttons.update({ f'{k}_price':  urwid.RadioButton(group, ('streak', u'Best'), state="first True") })
            self.buttons.update({ f'{k}_date':  urwid.RadioButton(group, ('streak', u'Last'), state="first True") })
            for button in group:
                urwid.connect_signal(button, 'postchange', lambda *args: self.update_historic_prices())
        
        urwid.connect_signal(self.buttons['clear'], 'click', lambda x: self.clear().update())
        urwid.connect_signal(self.buttons['exit'], 'click', lambda x: show_or_exit('esc'))
        
        dbview, spread, rating, marker = [ urwid.Text('') for i in range(0,4) ]
        
        self.text_fields.update({
            'dbview': dbview,
            'spread': spread,
            'rating': rating,
            'marker': marker,
        })
        
        self.clear()
        
        header = urwid.Text(u'Price Check', 'center')
        _copyright = urwid.Text(COPYRIGHT, 'center')
        
        banner = urwid.Pile([
            urwid.Padding(header, 'center', width=('relative', 100)),
            urwid.Padding(_copyright, 'center', width=('relative', 100)),
        ])
        banner = urwid.AttrMap(banner, 'banner')
        fields = dict([
            (k, urwid.LineBox(urwid.AttrMap(
                AutoCompletePopUp(
                    self.edit_fields[k],
                    self.apply_choice,
                    lambda: activity_manager.show(self.update())
                ), 'streak'), title=k.title(), title_align='left')
            ) for k in self.edit_fields
        ])
        fields['organic'] = urwid.AttrMap(self.organic_checkbox, 'bg')
        fields.update({
            'dbview': urwid.LineBox(
                urwid.AttrMap(dbview, 'streak'),
                title="Historic Prices",
                title_align='center',
            ),
            'spread': spread,
            'rating': rating,
            'marker': marker,
        })

        right_pane_widget = (16, urwid.Pile(map(
            lambda x: fields[x] if x is not None else urwid.Divider(),
            self.right_pane
        )))
        left_pane_widget = (urwid.Pile(map(
            lambda x: fields[x] if x is not None else urwid.Divider(),
            self.left_pane,
        )))
        left_pane_widget = (urwid.Pile([
            left_pane_widget,
            urwid.LineBox(urwid.AttrMap(
                urwid.Pile(map(
                    lambda x: fields[x] if x is not None else urwid.Divider(),
                    self.badge,
                )), 'badge',),
                title="Current Price",
                title_align='left',
            )
        ]))
        
        widget = urwid.Pile([
            banner,
            urwid.Divider(),
            urwid.Columns(
                [
                    (9, urwid.Pile([
                        urwid.Divider(),
                        urwid.AttrMap(self.buttons['clear'], 'streak'),
                        urwid.Divider(),
                    ])),
                    urwid.LineBox(
                        urwid.Columns([ v for k,v in self.buttons.items() if 'sort' in k]),
                        title="Sort price by",
                        title_align='left',
                    ),
                    (9, urwid.Pile([
                        urwid.Divider(),
                        urwid.AttrMap(self.buttons['exit'], 'streak'),
                        urwid.Divider(),
                    ]))
                ],
                dividechars=1,
            ),
            urwid.Columns((left_pane_widget, right_pane_widget),
                dividechars=0,
            ),
            *[ fields[c] if c is not None else urwid.Divider() for c in self.bottom_pane ],
        ])
        widget = urwid.Filler(widget, 'top') 
        widget = urwid.AttrMap(widget, 'bg')
        widget.original_widget.original_widget.set_focus_path([3,0,0])
        super().__init__(widget, 4, [3,0,0,0], [
                [0, 0,], [0,1,],
                [1,],
                [2,],
                [2,0,],
                [2,0,0,], [2,0,2], [2,2,0], [2,2,2],
                [3,0,], [3,0,0,],
                [3,0,1,0,], [3,0,1,1,], [3,0,1,2,],
                [3,1,3,],
                [4,],
            ]
        )
